/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: RemoteLauncher.java 1941 2007-10-14 17:57:14Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.osgi.plugins.launcher;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.net.URL;

import org.codehaus.plexus.util.cli.CommandLineException;
import org.codehaus.plexus.util.cli.Commandline;
import org.ow2.easybeans.osgi.plugins.launcher.api.LauncherException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * @author Florent BENOIT
 */
public class RemoteLauncher extends AbsLauncher {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(RemoteLauncher.class);

    /**
     * URLs of the classpath of the JVM to start.
     */
    private URL[] classpathURLs = null;

    /**
     * Process of the remote JVM.
     */
    private Process process = null;

    /**
     * Build a launcher with the given configuration object and the classpath
     * URLs used to create the forked JVM.
     * @param classpathURLs the URLs to use for the classpath of the JVM.
     * @param configuration the given configuration
     */
    public RemoteLauncher(final URL[] classpathURLs, final Configuration configuration) {
        super(configuration);
        this.classpathURLs = classpathURLs;
    }

    /**
     * Create a new command line that will launch a Bootstrap object on a remote JVM.
     * @throws LauncherException if the starting fails.
     */
    public void start() throws LauncherException {

        // Create command line
        Commandline commandline = new Commandline();

        // Sets the java command to use
        commandline.setExecutable(System.getProperty("java.home") + File.separator + "bin" + File.separator + "java");

        // Starting...
        writeInstruction(Bootstrap.STARTED_INSTRUCTION);

        // Build classpath
        String classpath = null;
        StringBuilder classpathBuStringBuilder = new StringBuilder();
        for (URL path : classpathURLs) {
            if (classpathBuStringBuilder.length() > 0) {
                classpathBuStringBuilder.append(File.pathSeparatorChar);
            }
            classpathBuStringBuilder.append(path.getPath());
        }
        classpath = classpathBuStringBuilder.toString();

        // Wait has to be done here and not in a remote process
        long waitAfterStart = getConfiguration().getWaitAfterStart();
        if (waitAfterStart > 0) {
            getConfiguration().setWaitAfterStart(0L);
        }

        // Dump the configuration in a file
        File configFile = null;
        try {
            configFile = File.createTempFile(System.getProperty("user.name"), ".configuration");
        } catch (IOException e) {
            throw new LauncherException("Cannot create temp file", e);
        }
        FileOutputStream fileOutputStream = null;
        ObjectOutputStream objectOutputStream = null;
        try {
            fileOutputStream = new FileOutputStream(configFile);
            objectOutputStream = new ObjectOutputStream(fileOutputStream);
            objectOutputStream.writeObject(getConfiguration());
        } catch (FileNotFoundException e) {
            throw new LauncherException("Cannot write the configuration object in the file '" + configFile + "'.", e);
        } catch (IOException e) {
            throw new LauncherException("Cannot write the configuration object in the file '" + configFile + "'.", e);
        } finally {
            // Close streams
            if (fileOutputStream != null) {
                try {
                    fileOutputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close the stream", e);
                }
            }
            if (objectOutputStream != null) {
                try {
                    objectOutputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close the stream", e);
                }
            }
        }

        // Sets the classpath
        commandline.createArgument().setValue("-classpath");
        commandline.createArgument().setValue(classpath);

        // Class to launch
        commandline.createArgument().setValue(Bootstrap.class.getName());

        // Path to the configuration file with a system property
        commandline.createArgument().setValue(Bootstrap.CONFIGURATION_PATH);
        commandline.createArgument().setValue(configFile.getAbsolutePath());

        logger.info("Launching the Remote JVM...");
        try {
            process = commandline.execute();
        } catch (CommandLineException e) {
            throw new LauncherException("Cannot execute the command line", e);
        }
        logger.info("Remote JVM started");

        // Trace the start execution
        InputStream error = process.getErrorStream();
        InputStream out = process.getInputStream();
        Thread errThread = new Thread(new ProcessLog(error, true));
        Thread outThread = new Thread(new ProcessLog(out, false));
        outThread.start();
        errThread.start();

        // Waiting is done here and not in the remote process
        logger.info("Waiting for " + (waitAfterStart + waitAfterStart));
        try {
            Thread.sleep((waitAfterStart + waitAfterStart) * SECONDS);
        } catch (InterruptedException e) {
            throw new LauncherException("Cannot wait", e);
        }

    }

    /**
     * Write a stop instruction in a file in order to tell to the remote JVM that the remote server needs to be stopped.
     * @throws LauncherException if the call is not done.
     */
    public void stop() throws LauncherException {
        writeInstruction(Bootstrap.STOPPED_INSTRUCTION);
        // Waiting for shutdown
        logger.info("Waiting for shutdown");
        try {
            Thread.sleep(getConfiguration().getWaitAfterStop() * SECONDS);
        } catch (InterruptedException e) {
            throw new LauncherException("Cannot wait", e);
        }
    }

    /**
     * Writes the given instruction in a monitored file.
     * @param instruction the instruction to write
     * @throws LauncherException if the instruction is not written
     */
    protected void writeInstruction(final String instruction) throws LauncherException {
        // Write in the monitored file
        FileWriter fileWriter = null;
        try {
            logger.info("Writing {0} in {1}", instruction, Bootstrap.MONITORED_FILE);
            fileWriter = new FileWriter(Bootstrap.MONITORED_FILE);
            fileWriter.write(instruction);
        } catch (IOException e) {
            throw new LauncherException("Cannot write in file '" + Bootstrap.MONITORED_FILE + "'", e);
        } finally {
            if (fileWriter != null) {
                try {
                    fileWriter.close();
                } catch (IOException e) {
                    logger.error("Cannot close file writer", e);
                }
            }
        }

    }

}
